import abc


class Transform(metaclass=abc.ABCMeta):
    def _repr_model_(self):
        """Return as representation of the object suitable for storing in mongo.

        This function returns a string representation of the object
        that is stored in a girder Job Model's 'args' or 'kwargs'
        fields.  It is for display purposes only.

        """
        return str(self)

    def cleanup(self):
        pass

    @abc.abstractmethod
    def transform(self):
        pass


class ResultTransform(Transform, metaclass=abc.ABCMeta):
    # This will be populated by the worker via the girder_before_task_publish hook,
    # and it will point to the job that was created for this task. If no job was created
    # for this task, this will remain `None`.
    job = None

    @abc.abstractmethod
    def transform(self, data):
        pass

    def exception(self):
        pass



from .girder import Girder


class GirderFolder(Girder):
    """Define a parameter representing a girder folder id."""

    def describe(self, **kwargs):
        desc = super().describe(**kwargs)
        desc['type'] = 'directory'
        desc['description'] = self.help or 'Select a folder'
        return desc



import os

accept_content = ['json', 'pickle', 'yaml', 'girder_io']

broker_url = os.environ.get(
    'GIRDER_WORKER_BROKER', 'amqp://guest:guest@localhost/')

result_backend = os.environ.get(
    'GIRDER_WORKER_BACKEND', 'rpc://guest:guest@localhost/')



import pytest

from girder_worker_utils import types


def test_serialize_boolean():
    arg = types.Boolean('arg')
    assert arg.serialize(None) is False
    assert arg.serialize(1) is True


def test_validate_choice_value():
    arg = types.StringMultichoice('arg')
    with pytest.raises(TypeError):
        arg.validate('not a list')


def test_serialize_integer():
    arg = types.Integer('arg')
    assert arg.serialize(1.0) == 1


def test_validate_integer_range():
    arg = types.Integer('arg', min=0, max=10)
    with pytest.raises(ValueError):
        arg.validate(-1)

    with pytest.raises(ValueError):
        arg.validate(11)


def test_validate_string():
    arg = types.String('arg')
    with pytest.raises(TypeError):
        arg.validate(1)


def test_validate_number_vector():
    arg = types.NumberVector('arg')
    with pytest.raises(TypeError):
        arg.validate('not a list')



from .choice import Choice


class NumberMultichoice(Choice):
    """Define a multichose numeric parameter type.

    Values of this type are iterable sequences of numbers
    all of which must be an element of a predefined set.

    >>> @argument('images', types.NumberMultichoice, choices=(5, 10, 15))
    ... def func(images=(5, 10)):
    ...     pass
    """

    paramType = 'number-enumeration-multiple'
    multiple = True



import mimetypes
import os
import shutil
import tempfile

from girder_client import GirderClient
from urllib.parse import urlencode

from ..transform import ResultTransform, Transform


class GirderClientTransform(Transform):
    def __init__(self, *args, **kwargs):  # noqa
        gc = kwargs.pop('gc', None)

        try:
            if gc is None:
                # We need to resolve Girder's API URL, but girder_worker can
                # specify a different value than what Girder gets from a rest
                # request.
                from girder_plugin_worker.utils import getWorkerApiUrl
                self.gc = GirderClient(apiUrl=getWorkerApiUrl())
                from girder.api.rest import getCurrentUser
                if getCurrentUser():
                    from girder.constants import TokenScope
                    from girder.models.token import Token
                    scope = [TokenScope.DATA_READ, TokenScope.DATA_WRITE]
                    if 'GIRDER_WORKER_JOB_GC_SCOPE' in os.environ:
                        try:
                            scope = list({
                                part.strip('"[\'] ') for part in
                                os.environ['GIRDER_WORKER_JOB_GC_SCOPE'].split(',')} | set(scope))
                        except Exception:
                            pass
                    token = Token().createToken(
                        days=7,
                        scope=scope,
                        user=getCurrentUser(),
                    )['_id']
                else:
                    from girder.api.rest import getCurrentToken
                    token = getCurrentToken()['_id']
                self.gc.token = token
            else:
                self.gc = gc
        except ImportError:
            self.gc = None


class GirderClientResultTransform(ResultTransform, GirderClientTransform):
    pass


class GirderFileId(GirderClientTransform):
    """
    This transform downloads a Girder File to the local machine and passes its
    local path into the function.

    :param _id: The ID of the file to download.
    :type _id: str
    """

    def __init__(self, _id, **kwargs):
        super().__init__(**kwargs)
        self.file_id = _id

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.file_id}')"

    def transform(self):
        self.file_path = os.path.join(
            tempfile.mkdtemp(), f'{self.file_id}')

        self.gc.downloadFile(self.file_id, self.file_path)

        return self.file_path

    def cleanup(self):
        if hasattr(self, 'file_path'):
            shutil.rmtree(os.path.dirname(self.file_path),
                          ignore_errors=True)


class GirderItemId(GirderClientTransform):
    """
    This transform downloads a Girder Item to a directory on the local machine
    and passes its local path into the function.

    :param _id: The ID of the item to download.
    :type _id: str
    """

    def __init__(self, _id, **kwargs):
        super().__init__(**kwargs)
        self.item_id = _id

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.item_id}')"

    def transform(self):
        temp_dir = tempfile.mkdtemp()
        self.item_path = os.path.join(temp_dir, self.item_id)
        # os.mkdir(self.item_path)

        self.gc.downloadItem(self.item_id, temp_dir, self.item_id)

        return self.item_path

    def cleanup(self):
        if hasattr(self, 'item_path'):
            shutil.rmtree(os.path.dirname(self.item_path),
                          ignore_errors=True)


class GirderItemMetadata(GirderClientTransform):
    def __init__(self, _id, **kwargs):
        super().__init__(**kwargs)
        self.item_id = _id

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.item_id}')"

    def transform(self, data):
        self.gc.addMetadataToItem(self.item_id, data)

        return data


class GirderUploadToItem(GirderClientResultTransform):
    """
    This is a result hook transform that uploads a file or flat directory of files
    to an item in Girder.

    :param _id: The ID of the item to upload into.
    :type _id: str
    :param delete_file: Whether to delete the local data afterward
    :type delete_file: bool
    :param upload_kwargs: Additional kwargs to pass to the upload method.
    :type upload_kwargs: dict
    """

    def __init__(self, _id, delete_file=False, upload_kwargs=None, **kwargs):
        super().__init__(**kwargs)
        self.item_id = _id
        self.upload_kwargs = upload_kwargs or {}
        self.delete_file = delete_file

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.item_id}')"

    def transform(self, path):
        self.output_file_path = path
        if os.path.isdir(path):
            for f in os.listdir(path):
                f = os.path.join(path, f)
                if os.path.isfile(f):
                    self.gc.uploadFileToItem(self.item_id, f, **self.upload_kwargs)
        else:
            self.gc.uploadFileToItem(self.item_id, path, **self.upload_kwargs)
        return self.item_id

    def cleanup(self):
        if self.delete_file is True and hasattr(self, 'output_file_path'):
            if os.path.isdir(self.output_file_path):
                shutil.rmtree(self.output_file_path)
            else:
                os.remove(self.output_file_path)


class GirderUploadToFolder(GirderClientResultTransform):
    """
    This is a result hook transform that uploads a file or directory recursively
    to a folder in Girder.

    :param _id: The ID of the folder to upload into.
    :type _id: str
    :param delete_file: Whether to delete the local data afterward
    :type delete_file: bool
    :param upload_kwargs: Additional kwargs to pass to the upload method.
    :type upload_kwargs: dict
    """

    def __init__(self, _id, delete_file=False, upload_kwargs=None, must_exist=True, **kwargs):
        super().__init__(**kwargs)
        self.folder_id = _id
        self.upload_kwargs = upload_kwargs or {}
        self.delete_file = delete_file
        self.must_exist = True

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.folder_id}')"

    def _uploadFolder(self, path, folder_id):
        for f in os.listdir(path):
            fpath = os.path.join(path, f)
            if os.path.isfile(fpath):
                self.gc.uploadFileToFolder(folder_id, fpath, **self.upload_kwargs)
            elif os.path.isdir(fpath) and not os.path.islink(fpath):
                folder = self.gc.createFolder(folder_id, f, reuseExisting=True)
                self._uploadFolder(fpath, folder['_id'])

    def transform(self, path):
        self.output_file_path = path
        if not self.must_exist and not os.path.exists(path):
            return None
        if os.path.isdir(path):
            self._uploadFolder(path, self.folder_id)
        else:
            self.gc.uploadFileToFolder(self.folder_id, path, **self.upload_kwargs)
        return self.folder_id

    def cleanup(self):
        if self.delete_file is True and hasattr(self, 'output_file_path'):
            if os.path.isdir(self.output_file_path):
                shutil.rmtree(self.output_file_path)
            else:
                os.remove(self.output_file_path)


class GirderUploadJobArtifact(GirderClientResultTransform):
    """
    This class can be used to upload a directory of files or a single file
    as artifacts attached to a Girder job. These files are only uploaded
    if they exist, so this is an optional output.

    Currently, only a flat directory of files is supported; the transform does not
    recurse through nested directories, though that may change in the future.

    :param job_id: The ID of the job to attach the file(s) to.
    :type job_id: str or ObjectId
    :param name: Name for the artifact (only if it's a single file).
    :type name: str
    """

    def __init__(self, job_id=None, name=None, **kwargs):
        super().__init__(**kwargs)
        self.job_id = job_id
        self.name = name

    def _repr_model_(self):
        return f"{self.__class__.__name__}('{self.job_id}')"

    def _upload_artifact(self, file, name=None):
        qs = urlencode({
            'name': name or os.path.basename(file),
            'size': os.stat(file).st_size,
            'mimeType': mimetypes.guess_type(file)[0]
        })
        with open(file, 'rb') as fh:
            self.gc.post('job/%s/artifact?%s' % (self.job_id, qs), data=fh)

    def transform(self, path):
        if self.job_id is None:
            self.job_id = str(self.job['_id'])

        if os.path.isdir(path):
            for f in os.listdir(path):
                f = os.path.join(path, f)
                if os.path.isfile(f):
                    self._upload_artifact(f)
        elif os.path.isfile(path):
            self._upload_artifact(path, self.name)



import numbers

from .base import Base


class Number(Base):
    """Define a numeric parameter type optionally in a given range.

    Values accepted by this parameter can be any numeric value.
    If min/max/step are provided, then the values must be a float
    or int.

    >>> @argument('value', types.Number, min=10, max=100, step=10)
    ... def func(value):
    ...     pass
    """

    paramType = 'number'

    def __init__(self, *args, **kwargs):
        """Construct a new numeric parameter type.

        :param float min: The minimum valid value
        :param float max: The maximum valid value
        :param float step: The resolution of valid values
        """
        super().__init__(*args, **kwargs)
        self.min = kwargs.get('min')
        self.max = kwargs.get('max')
        self.step = kwargs.get('step')

    def describe(self, **kwargs):
        desc = super().describe(**kwargs)

        if self.min is not None:
            desc['min'] = self.min
        if self.max is not None:
            desc['max'] = self.max
        if self.step is not None:
            desc['step'] = self.step

        desc['type'] = self.paramType
        desc['description'] = self.help or 'Select a number'
        return desc

    def validate(self, value):
        if not isinstance(value, numbers.Number):
            raise TypeError('Expected a number for parameter "%s"' % self.name)

        if self.min is not None and value < self.min:
            raise ValueError('Expected %s <= %s' % (str(self.min), str(value)))

        if self.max is not None and value > self.max:
            raise ValueError('Expected %s >= %s' % (str(self.max), str(value)))

    def serialize(self, value):
        if self.step is not None:
            n = round(float(value) / self.step)
            value = n * self.step
        return value

    def deserialize(self, value):
        try:
            value = float(value)
        except ValueError:
            pass
        return value



from .base import Base


class Choice(Base):
    """Define base functionality for multi choice parameter."""

    #: The explicit type provided by the subclass
    paramType = None

    #: Whether the parameter accepts multiple values from the choices
    multiple = False

    def __init__(self, *args, **kwargs):
        """Construct a choice parameter.

        :param list choices: A list of valid values.
        """
        self.choices = kwargs.get('choices', [])
        super().__init__(*args, **kwargs)

    def describe(self, *args, **kwargs):
        desc = super().describe(*args, **kwargs)
        desc['type'] = self.paramType
        desc['values'] = self.choices
        desc['description'] = 'Choose from a list of values'
        return desc

    def _validate_one(self, value):
        if value not in self.choices:
            raise ValueError('Invalid value provided for "%s"' % self.name)

    def validate(self, value):
        if not self.multiple:
            value = [value]

        if not isinstance(value, (list, tuple)):
            raise TypeError(
                'Expected a list or tuple for "%s"' % self.name)
        for v in value:
            self._validate_one(v)



from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

import time

from .tee import Tee, tee_stderr, tee_stdout

import requests
from requests import HTTPError

# Disable urllib3 warnings about certificate validation. As they are printed in the console, the
# messages are sent to Girder, creating an infinite loop.
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def _walk_obj(obj, func, leaf_condition=None):
    """Walk through a nested object applying func to each leaf element.

    This function returns a recursively built structure from a nested
    tree of simple container types (e.g. dict, list, tuple) by
    applying func to each leaf node in the tree. By default a leaf
    node is considered to be any object that is not a dict, list or
    tuple.

    leaf_condition may be used if certain types of lists, tuples or
    dicts should be considered leaf nodes. leaf_condition should be
    passed a function that takes a sub-tree and returns True if that
    sub-tree is a leaf node, or False if _walk_obj should continue to
    descend through the sub-tree.

    """
    if callable(leaf_condition):
        if leaf_condition(obj):
            return func(obj)

    if isinstance(obj, dict):
        return {k: _walk_obj(v, func, leaf_condition=leaf_condition)
                for k, v in obj.items()}

    elif isinstance(obj, list):
        return [_walk_obj(v, func, leaf_condition=leaf_condition) for v in obj]

    elif isinstance(obj, tuple):
        return tuple(_walk_obj(list(obj), func, leaf_condition=leaf_condition))

    return func(obj)


BUILTIN_CELERY_TASKS = [
    'celery.accumulate'
    'celery.backend_cleanup',
    'celery.chain',
    'celery.chord',
    'celery.chord_unlock',
    'celery.chunks',
    'celery.group',
    'celery.map',
    'celery.starmap']


def is_builtin_celery_task(task):
    return task in BUILTIN_CELERY_TASKS


def _maybe_model_repr(obj):
    if hasattr(obj, '_repr_model_') and callable(obj._repr_model_):
        return obj._repr_model_()
    return obj


# Access to the correct "Inspect" instance for this worker
_inspector = None


def _worker_inspector(task):
    global _inspector

    if _inspector is None:
        try:
            # Celery >= 5
            from ..app import app
            inspect = app.control.inspect
        except AttributeError:
            # Celecy < 5
            from celery.app.control import Inspect as inspect  # noqa: N813
        _inspector = inspect([task.request.hostname])

    return _inspector


# Get this list of currently revoked tasks for this worker
def _revoked_tasks(task):
    _revoked = _worker_inspector(task).revoked()

    if _revoked is None:
        return []

    return _revoked.get(task.request.hostname, [])


def deserialize_job_info_spec(**kwargs):
    return JobManager(**kwargs)


class JobSpecNotFound(Exception):
    pass


def _job_manager(request=None, headers=None, kwargs=None):

    girder_client_session_kwargs = {}
    if hasattr(request, 'girder_client_session_kwargs'):
        girder_client_session_kwargs = request.girder_client_session_kwargs

    if hasattr(request, 'jobInfoSpec'):
        jobSpec = request.jobInfoSpec

    # We are being called from revoked signal
    elif headers is not None and \
            'jobInfoSpec' in headers:
        jobSpec = headers['jobInfoSpec']

    # Deprecated: This method of passing job information
    # to girder_worker is deprecated. Newer versions of girder
    # pass this information automatically as apart of the
    # header metadata in the worker scheduler.
    elif kwargs and 'jobInfo' in kwargs:
        jobSpec = kwargs.pop('jobInfo', {})

    else:
        raise JobSpecNotFound

    return deserialize_job_info_spec(
        **jobSpec, girder_client_session_kwargs=girder_client_session_kwargs)


def _update_status(task, status):
    task.job_manager.updateStatus(status)


def is_revoked(task):
    """
    Utility function to check if a task has been revoked.

    Eager tasks are never considered revoked.

    :param task: The task.
    :type task: celery.app.task.Task
    :return: True, if this task is in the revoked list for this worker, False
            otherwise.
    :rtype: bool
    """
    if task.request.is_eager:
        return False
    return task.request.id in _revoked_tasks(task)


def girder_job(title=None, type='celery', public=False,
               handler='celery_handler', otherFields=None):
    """Decorator that populates a girder_worker celery task with
    girder's job metadata.

    :param title: The title of the job in girder.
    :type title: str
    :param type: The type of the job in girder.
    :type type: str
    :param public: Public read access flag for girder.
    :type public: bool
    :param handler: If this job should be handled by a specific handler,
        'celery_handler' by default cannot be scheduled in girder.
    :param otherFields: Any additional fields to set on the job in girder.
    :type otherFields: dict
    """
    otherFields = otherFields or {}

    def _girder_job(task_obj):
        task_obj._girder_job_title = title
        task_obj._girder_job_type = type
        task_obj._girder_job_public = public
        task_obj._girder_job_handler = handler
        task_obj._girder_job_other_fields = otherFields
        return task_obj

    return _girder_job


class JobStatus:
    INACTIVE = 0
    QUEUED = 1
    RUNNING = 2
    SUCCESS = 3
    ERROR = 4
    CANCELED = 5

    FETCHING_INPUT = 820
    CONVERTING_INPUT = 821
    CONVERTING_OUTPUT = 822
    PUSHING_OUTPUT = 823
    CANCELING = 824


class StateTransitionException(Exception):
    pass


class TeeCustomWrite(Tee):
    def __init__(self, func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._write_func = func

    def write(self, *args, **kwargs):
        self._write_func(*args, **kwargs)
        super().write(*args, **kwargs)


@tee_stdout
class TeeStdOutCustomWrite(TeeCustomWrite):
    pass


@tee_stderr
class TeeStdErrCustomWrite(TeeCustomWrite):
    pass


class JobManager:
    """
    This class can be used to write log messages to Girder by capturing
    stdout/stderr printed within the context and sending them in a
    rate-limited manner to Girder. This is not threadsafe since it changes
    the global values of sys.stdout/sys.stderr.

    It also exposes utilities for updating other job fields such as progress
    and status.
    """

    def __init__(self, logPrint, url, method=None, headers=None, interval=0.5,
                 reference=None, girder_client_session_kwargs=None):
        """
        :param on: Whether print messages should be logged to the job log.
        :type on: bool
        :param url: The job update URL.
        :param method: The HTTP method to use when updating the job.
        :param headers: Optional HTTP header dict
        :param interval: Minimum time interval at which to send log updates
        back to Girder over HTTP (seconds).
        :type interval: int or float
        :param reference: optional reference to store with the job.
        """
        self.logPrint = logPrint
        self.method = method or 'PUT'
        self.url = url
        self.headers = headers or {}
        self.interval = interval
        self.status = None
        self.reference = reference

        self._last = time.time()
        self._buf = b''
        self._progressTotal = None
        self._progressCurrent = None
        self._progressMessage = None

        self._session = requests.Session()
        retryAdapter = requests.adapters.HTTPAdapter(max_retries=10)
        self._session.mount('http://', retryAdapter)
        self._session.mount('https://', retryAdapter)

        if girder_client_session_kwargs:
            for attr, value in girder_client_session_kwargs.items():
                setattr(self._session, attr, value)

        if logPrint:
            self._stdout = TeeStdOutCustomWrite(self.write)
            self._stderr = TeeStdErrCustomWrite(self.write)

    def cleanup(self):
        self._session.close()
        if self.logPrint:
            self._stdout.reset()
            self._stderr.reset()

    def _flush(self):
        """
        If there are contents in the buffer, send them up to the server. If the
        buffer is empty, this is a no-op.
        """
        if not self.url:
            return

        if len(self._buf) or self._progressTotal or self._progressMessage or \
                self._progressCurrent is not None:
            data = {
                'progressTotal': self._progressTotal,
                'progressCurrent': self._progressCurrent,
                'progressMessage': self._progressMessage
            }
            if self._buf:
                data['log'] = self._buf

            req = self._session.request(
                self.method.upper(), self.url, allow_redirects=True,
                headers=self.headers, data=data)
            req.raise_for_status()
            self._buf = b''

    def write(self, message, forceFlush=False):
        """
        Append a message to the log for this job. If logPrint is enabled, this
        will be called whenever stdout or stderr is printed to. Otherwise it
        can be called manually and will still perform rate-limited flushing to
        the server.

        :param message: The message to append to the job log.
        :type message: str
        :param forceFlush: Whether to force the write of this event to the
            server. Useful if you don't expect another update for some time.
        :type forceFlush: bool
        """
        if isinstance(message, str):
            message = message.encode('utf8')

        self._buf += message

        if forceFlush or time.time() - self._last > self.interval:
            self._flush()
            self._last = time.time()

    def updateStatus(self, status):
        """
        Update the status field of a job.

        :param status: The status to set on the job.
        :type status: JobStatus
        """
        if not self.url or status is None or status == self.status:
            return

        # Ensure that the logs are flushed before the status is changed
        self._flush()
        self.status = status
        try:
            req = self._session.request(self.method.upper(), self.url, headers=self.headers,
                                        data={'status': status}, allow_redirects=True)
            req.raise_for_status()
        except HTTPError as hex:
            if hex.response.status_code == 400:
                json_response = hex.response.json()
                if 'field' in json_response and json_response['field'] == 'status':
                    print(json_response['message'])
                    raise StateTransitionException(json_response['message'], hex)
                else:
                    raise
            else:
                raise

    def updateProgress(self, total=None, current=None, message=None,
                       forceFlush=False):
        """
        Update the progress information about a job.

        :param total: The total progress value, or None to leave the same.
        :type total: int, float, or None
        :param current: The current progress value, or None to leave the same.
        :type current: int, float, or None
        :param message: Progress message to set, or None to leave the same.
        :type message: str or None
        :param forceFlush: Whether to force the write of this event to the
            server. Useful if you don't expect another update for some time.
        :type forceFlush: bool
        """
        if total is not None:
            self._progressTotal = total
        if current is not None:
            self._progressCurrent = current
        if message is not None:
            self._progressMessage = message

        if forceFlush or time.time() - self._last > self.interval:
            self._flush()
            self._last = time.time()

    def refreshStatus(self):
        """
        Refresh the status field from Girder
        """
        r = self._session.get(self.url, headers=self.headers, allow_redirects=True)
        self.status = r.json()['status']

        return self.status



import girder_client
from unittest import mock
import os
import pytest

from girder_worker_utils.transforms.contrib import girder_io


@pytest.fixture
def mock_gc():
    return mock.MagicMock(spec=girder_client.GirderClient)


@pytest.fixture
def mock_rmtree():
    with mock.patch('shutil.rmtree') as rmtree:
        yield rmtree


def test_GirderFileIdAllowDirect_without_env(mock_gc, mock_rmtree):
    local_path = os.path.abspath(__file__)
    t = girder_io.GirderFileIdAllowDirect('the_id', 'the_name', local_path, gc=mock_gc)
    t.transform()
    mock_gc.downloadFile.assert_called_once()
    assert 'the_id' in mock_gc.downloadFile.call_args[0]
    mock_rmtree.assert_not_called()
    t.cleanup()
    mock_rmtree.assert_called_once()


@mock.patch.dict(os.environ, {'GW_DIRECT_PATHS': 'true'})
def test_GirderFileIdAllowDirect_with_env(mock_gc, mock_rmtree):
    local_path = os.path.abspath(__file__)
    t = girder_io.GirderFileIdAllowDirect('the_id', 'the_name', local_path, gc=mock_gc)
    t.transform()
    mock_gc.downloadFile.assert_not_called()
    t.cleanup()
    mock_rmtree.assert_not_called()


@mock.patch.dict(os.environ, {'GW_DIRECT_PATHS': 'true'})
def test_GirderFileIdAllowDirect_with_env_and_unreachable_file(mock_gc, mock_rmtree):
    t = girder_io.GirderFileIdAllowDirect('the_id', 'the_name', 'the_path', gc=mock_gc)
    t.transform()
    mock_gc.downloadFile.assert_called_once()
    assert 'the_id' in mock_gc.downloadFile.call_args[0]
    mock_rmtree.assert_not_called()
    t.cleanup()
    mock_rmtree.assert_called_once()



import jsonpickle
import pytest


from girder_worker_utils.transform import Transform


class CaptureTransform(Transform):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def transform(self):
        return self.args, self.kwargs


class BasicCustomTransform(Transform):
    def __init__(self, arg1, kwarg1=None):
        self.arg1 = arg1
        self.kwarg1 = kwarg1

    def transform(self):
        return None


class ComputedCustomTransform(Transform):
    def __init__(self, kwarg1=None):
        if kwarg1 is None:
            self.kwarg1 = 'COMPUTED VALUE'

    def transform(self):
        return None


class CustomSerializeTransform(Transform):
    def __init__(self, arg1, kwarg1='some_value'):
        self.arg1 = arg1
        self.kwarg1 = kwarg1

    def __getstate__(self):
        return {'some_attr': 'some_value'}

    def transform(self):
        return None


def test_instantiate_transform_behavior():
    with pytest.raises(TypeError):
        Transform()


def test_basic_transform_behavior():
    assert CaptureTransform().transform() == ((), {})


def test_noop_transform():
    assert CaptureTransform('some value').transform() == (('some value', ), {})


def test_non_hook_transform_roundtrip():
    obj = {'key': 'value',
           'key2': ['list', 'of', 'values'],
           'key3': {'inner': 'dict'}}

    ret = jsonpickle.decode(jsonpickle.encode(obj))
    assert ret == obj


def test_transform_serialize_roundtrip():
    args = ('arg1', 'arg2', 'arg3')
    kwargs = {'kwarg1': 'kwarg1', 'kwarg2': 'kwarg2'}
    original_instance = CaptureTransform(*args, **kwargs)
    new_instance = jsonpickle.decode(jsonpickle.encode(original_instance))

    assert isinstance(new_instance, Transform)
    assert new_instance.args == args
    assert new_instance.kwargs == kwargs


def test_basic_custom_transform_arg1_roundtrip():
    new_instance = jsonpickle.decode(jsonpickle.encode(
        BasicCustomTransform('some_value')
    ))

    assert isinstance(new_instance, BasicCustomTransform)
    assert new_instance.arg1 == 'some_value'


def test_basic_custom_transform_kwarg1_None_roundtrip():
    new_instance = jsonpickle.decode(jsonpickle.encode(
        BasicCustomTransform(None)
    ))

    assert isinstance(new_instance, BasicCustomTransform)
    assert new_instance.kwarg1 is None


def test_basic_custom_transform_kwarg1_not_None_roundtrip():
    new_instance = jsonpickle.decode(jsonpickle.encode(
        BasicCustomTransform(None, kwarg1='some_value')
    ))

    assert isinstance(new_instance, BasicCustomTransform)
    assert new_instance.kwarg1 == 'some_value'


def test_computed_custom_transform_kwarg1_has_computed_value_roundtrip():
    new_instance = jsonpickle.decode(jsonpickle.encode(
        ComputedCustomTransform()
    ))

    assert isinstance(new_instance, ComputedCustomTransform)
    assert new_instance.kwarg1 == 'COMPUTED VALUE'


def test_custom_serialize_transform_has_correct_attrs_roundtrip():
    new_instance = jsonpickle.decode(jsonpickle.encode(
        CustomSerializeTransform('arg1', 'kwarg1')
    ))

    assert isinstance(new_instance, CustomSerializeTransform)
    assert not hasattr(new_instance, 'arg1')
    assert not hasattr(new_instance, 'kwarg1')
    assert hasattr(new_instance, 'some_attr')
    assert new_instance.some_attr == 'some_value'



from .number import Number


class Range(Number):
    """Define numeric parameter with valid values in a given range.

    >>> @argument('value', types.Range, min=10, max=100, step=10)
    ... def func(value):
    ...     pass
    """

    paramType = 'range'



from .string import String
from .vector import Vector


class StringVector(Vector):
    """Define a parameter which takes a list of strings.

    >>> @argument('people', types.StringVector)
    ... def func(people=('alice', 'bob')):
    ...     pass
    """

    paramType = 'string-vector'
    elementClass = String



from girder_worker import GirderWorkerPluginABC


class BaseTestPlugin(GirderWorkerPluginABC):
    def __init__(self, *args, **kwargs):
        pass

    def task_imports(self):
        return []


class TestPlugin1(BaseTestPlugin):
    pass


class TestPlugin2(BaseTestPlugin):
    def task_imports(self):
        return ['girder_worker._test_plugins.tasks']


class TestPluginException1(BaseTestPlugin):
    def __init__(self, *arg, **kwargs):
        raise Exception('Exception in constructor')


class TestPluginException2(BaseTestPlugin):
    def task_imports(self):
        raise Exception('Exception in task_imports')


class TestPluginInvalidModule(BaseTestPlugin):
    def task_imports(self):
        return ['not a valid module']
